def create_app(test_config=None):
    import os, secrets
    from flask import Flask
    from ayta.extensions import limiter, caching, celery_init_app, oidc
    from werkzeug.middleware.proxy_fix import ProxyFix

    from . import filters

    config = {'MONGO_CONNECTION': os.environ.get('AYTA_MONGOCONNECTION', 'mongodb://root:example@192.168.66.140:27017'),
              'OIDC_PROVIDER': os.environ.get('AYTA_OIDC_PROVIDER', 'https://auth.ventilaar.nl'),
              'OIDC_ID': os.environ.get('AYTA_OIDC_ID', 'ayta'),
              'CACHE_DEFAULT_TIMEOUT': int(os.environ.get('AYTA_CACHETIMEOUT', 6)),
              'DEBUG': bool(os.environ.get('AYTA_DEBUG', False)),
              'DOMAIN': os.environ.get('AYTA_DOMAIN', 'https://testing.mashallah.nl'),
              'CELERY': {'broker_url': str(os.environ.get('AYTA_CELERYBROKER', 'amqp://guest:guest@192.168.66.140:5672/'))}
             }
    
    # Static Flask configuration options
    
    config['CELERY']['task_ignore_result'] = True
    config['CACHE_TYPE'] = 'SimpleCache'
    config['SECRET_KEY'] = secrets.token_bytes(32)
    
    # Celery Periodic tasks
    
    config['CELERY']['beat_schedule'] = {}
    config['CELERY']['beat_schedule']['Renew WebSub endpoints'] = {'task': 'ayta.tasks.websub_renew_expiring', 'schedule': 4000}
    config['CELERY']['beat_schedule']['Process WebSub data'] = {'task': 'ayta.tasks.websub_process_data', 'schedule': 100}
    
    app = Flask(__name__)
    app.config.from_mapping(config)
    
    limiter.init_app(app)
    caching.init_app(app)
    oidc.init_app(app)
    celery_init_app(app)

    app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1)
    
    app.jinja_env.filters['pretty_duration'] = filters.pretty_duration
    app.jinja_env.filters['pretty_time'] = filters.pretty_time
    app.jinja_env.filters['current_time'] = filters.current_time
    app.jinja_env.filters['epoch_time'] = filters.epoch_time

    from .blueprints import watch
    from .blueprints import index
    from .blueprints import admin
    from .blueprints import search
    from .blueprints import channel
    from .blueprints import auth
    from .blueprints import api

    app.register_blueprint(watch.bp)
    app.register_blueprint(index.bp)
    app.register_blueprint(admin.bp)
    app.register_blueprint(search.bp)
    app.register_blueprint(channel.bp)
    app.register_blueprint(auth.bp)
    app.register_blueprint(api.bp)

    return app