import json
import pymongo
import secrets
from bson.objectid import ObjectId

from flask import current_app, g

from .filters import current_time

##########################################
#              SETUP FLASK               #
##########################################

def get_nosql():
    """Connect to the application's configured database. The connection is unique for each request and will be reused if this is called again."""
    if "nosql" not in g:
        g.nosql = Mango(current_app.config["MONGO_CONNECTION"])

    return g.nosql

##########################################
#                 ORM                    #
##########################################
class Mango:
    def __init__(self, connect):
        from datetime import datetime, timedelta
        
        self.datetime = datetime
        self.timedelta = timedelta
        
        try:
            self.client = pymongo.MongoClient(connect)
        except ConnectionError:
            print('MongoDB connection error')
        
        self.db = self.client['ayta']
        self.channels = self.db['channels']
        self.info_json = self.db['info_json']
        self.download_queue = self.db['download_queue']
        self.run_log = self.db['run_log']
        self.channel_log = self.db['channel_log']
        self.websub_callbacks = self.db['websub_callbacks']
        self.websub_data = self.db['websub_data']
        self.reports = self.db['reports']
        self.queue_endpoints = self.db['queue_endpoints']
        self.users = self.db['users']
            
        self.ensure_indexes()
   
    def ensure_indexes(self):
        required = {
            'channels': [
                ('id_1', True)
            ],
            'info_json': [
                ('id_1', True),
                ('channel_id_1', False),
                ('uploader_1', False)
            ],
            'websub_callbacks': [
                ('id', True)
            ],
            'websub_data': [
                ('callback_id', False)
            ],
            'channel_log': [
                ('run_id', False),
                ('id', False)
            ]
        }

        # cursed but works
        for collection in self.db.list_collection_names():
            if collection in required:
                indexes = self.db[collection].index_information()
                for index in required[collection]:
                    if index[0] not in indexes:
                        self.db[collection].create_index(index[0], unique=index[1])
                        
    def clean_info_json(self):
        self.info_json.update_many({}, {'$unset': {'formats': 1, 
                                                   'thumbnails': 1,
                                                   'thumbnail': 1,
                                                   'http_headers': 1,
                                                   'automatic_captions': 1,
                                                   'subtitles': 1,
                                                   'url': 1,
                                                   '_filename': 1
                                                   }})
    
    ##########################################
    #             general functions          #
    ##########################################
    
    def gen_stats(self):
        """ Returns DICT; Channel statistics given the dict key """
        stats = {}
        
        stats['videos'] = self.info_json.count_documents({})
        stats['channels'] = self.channels.count_documents({})
        stats['queue'] = self.download_queue.count_documents({})
        
        return stats
            
    def search_videos(self, query):
        # search the index for the requested query. return limited keys
        results = self.info_json.find({"$text": {"$search": query}}, 
                                      {"score": {"$meta": "textScore"},
                                       "title": 1,
                                       "uploader": 1,
                                       "upload_date": 1,
                                       "id": 1,
                                       "channel_id": 1}
                                     )
        
        # sort the results based on highest score and limit final results to 20
        results = sorted(results, key=lambda x: x.get('score'), reverse=True)[:20]
        return tuple(results)
    
    ##########################################
    #              user operations           #
    ##########################################
    
    def list_all_users(self):
        return self.users.find({})
    
    def add_user(self, sub, alias, description=None):
        return self.users.insert_one({'sub': sub, 'alias': alias, 'description': description}).inserted_id
        
    def delete_user(self, sub):
        self.users.delete_one({'sub': sub})
        
    def get_user(self, sub):
        """ Returns True if sub exists, otherwise False """
        if self.users.count_documents({'sub': sub}) >= 1:
            return True
            
        return False
        
    
    ##########################################
    #            channel operations          #
    ##########################################
    
    def list_all_channels(self, active=False, websub=False):
        """ Returns a SET of YouTube channel ID's; Depending on given positional BOOL only active channels or everything"""
        search_terms = {}
        
        if active:
            search_terms['active'] = True
        elif websub:
            search_terms['websub'] = True
        
        channels = []
        for channel in self.channels.find(search_terms, {'id': 1}):
            channels.append(channel['id'])
        return tuple(channels)
        
    def get_channel_videos_count(self, channelId):
        """
        Returns int of total video count on specific channel
        """
        return self.info_json.count_documents({'channel_id': channelId})
        
    def get_channel_videoIds(self, channelId):
        ids = []
        for video in self.info_json.find({'channel_id': channelId}, {'id': 1}):
            ids.append(video['id'])
        return tuple(ids)
 
    def get_channel_info(self, channelId):
        return self.channels.find_one({'id': channelId})
        
        
    def update_channel_key(self, channelId, key, value):
        self.channels.update_one({'id': channelId}, {"$set": {key: value}})
        return True
    
    def insert_new_channel(self, channelId, originalName, addedDate):
        if self.channels.count_documents({'id': channelId}) >= 1:
            return False
    
        return self.channels.insert_one({'id': channelId, 'original_name': originalName, 'added_date': addedDate, 'active': False, 'websub': False}).inserted_id
    
    ##########################################
    #            video operations          #
    ##########################################
    
    def check_exists(self, vId):
        """ Returns BOOL; Given positional argument which is a valid or invalid YouTube video ID STR"""
        if self.info_json.count_documents({'id': vId}) >= 1:
            return True
        return False
        
    def get_orphaned_videos(self):
        """ Returns a SET of YouTube video ID's which have info_jsons in the collection but no permanent channel is defined. SLOW OPERATION """
        # Ok lemme explain. Perform inner join from channel collection on channel_id key. match only the fields which are empty. return video id
        pipeline = [{'$match': {'_status': 'available'}},
                    {'$lookup': {'from': 'channels', 'localField': 'channel_id', 'foreignField': 'id', 'as': 'channel'}},
                    {'$match': {'channel': {'$size': 0}}},{'$project': {'id': 1}},
                    {'$project': {'id': 1}}]
        
        results = self.info_json.aggregate(pipeline)
        ids = [result['id'] for result in results]

        return tuple(ids)
    
    def get_recent_videos(self, count=99):
        """ Returns a SET of YouTube video ID's which have been added last to the info_json collection """
        result = self.info_json.find({'_status': 'available'}, {'_id': 0, 'id': 1}, sort=[('_id', pymongo.DESCENDING)]).limit(count)

        ids = []
        
        for id in result:
            ids.append(id['id'])

        return tuple(ids)
        
    def get_video_info(self, videoId, limited=False):
        if limited:
            projection = {'_id': 1, 'id': 1, 'title': 1, 'upload_date': 1, 'description': 1, 'channel_id': 1, 'uploader': 1, 'epoch': 1, '_title_slug': 1}
        else:
            projection = {}
        
        return self.info_json.find_one({'id': videoId}, projection)
    
    ##########################################
    #          REPORTING FUNCTIONS           #
    ##########################################
        
    def insert_report(self, videoId, reason):
        if reason not in ['auto-video', 'metadata', 'illegal']:
            return False
    
        return self.reports.insert_one({'videoId': videoId, 'status': 'open', 'reason': reason, 'reporting_time': current_time(object=True)}).inserted_id
        
    def list_reports(self):
        reports = []
        for report in self.reports.find({}):
            reports.append(report)
        return tuple(reports)
        
    def close_report(self, _id):
        _id = ObjectId(_id)
        return self.reports.update_one({'_id': _id}, {'$set': {'status': 'closed', 'closing_time': current_time(object=True)}})

    ##########################################
    #           RUNLOG FUNCTIONS             #
    ##########################################
    
    def get_runs(self):
        return self.run_log.find({})
    
    def get_run(self, runId):
        run = self.run_log.find_one({'_id': ObjectId(runId)})
        run['channel_runs'] = list(self.channel_log.find({'run_id': ObjectId(runId)}))
        
        return run
        
    def clean_runs(self, keep=3):
        runs = list(self.run_log.find({}, {'_id': 1}).sort('time', pymongo.ASCENDING))
        
        for x in range(keep):
            if len(runs) == 0:
                return True
            runs.pop(-1)
        
        # runs is now a list of objectid's to remove
        
        for run in runs:
            self.channel_log.delete_many({'run_id': run['_id']})
            self.run_log.delete_one({'_id': run['_id']})
        
        return True
        
    ##########################################
    #           WEBSUB FUNCTIONS             #
    ##########################################
    
    def websub_newCallback(self, channelId):
        callbackId = secrets.token_urlsafe(16)
        
        self.websub_callbacks.insert_one({'id': callbackId, 'channel': channelId, 'status': 'new', 'created_time': current_time(object=True)})
        
        return callbackId

    def websub_requestingCallback(self, callbackId):
        return self.websub_callbacks.update_one({'id': callbackId}, {'$set': {'status': 'requesting', 'requesting_time': current_time(object=True)}})

    def websub_activateCallback(self, callbackId, lease):
        status = self.websub_callbacks.find_one({'id': callbackId}, {'status': 1})
        
        if not status:
            return False
            
        status = status.get('status')
        
        if status in ['requesting']:
            self.websub_callbacks.update_one({'id': callbackId}, {'$set': {'status': 'active', 'activation_time': current_time(object=True), 'lease': int(lease)}})
            return True
        
        return False

    def websub_existsCallback(self, callbackId, channel=False):
        if channel:
            query = {'channel': callbackId, 'status': {'$in': ['requesting', 'active', 'retiring']}}
        else:
            query = {'id': callbackId, 'status': {'$in': ['requesting', 'active', 'retiring']}}
            
        status = self.websub_callbacks.find_one(query, {'id': 1, 'status': 1})

        if status:
            return status.get('id')
        
        return False
        
    def websub_retiringCallback(self, callbackId):
        return self.websub_callbacks.update_one({'id': callbackId}, {'$set': {'status': 'retiring', 'retiring_time': current_time(object=True)}})
        
    def websub_retireCallback(self, callbackId):
        return self.websub_callbacks.update_one({'id': callbackId}, {'$set': {'status': 'retired', 'retired_time': current_time(object=True)}})
        
    def websub_deleteCallback(self, callbackId):
        return self.websub_callbacks.delete_one({'id': callbackId})
        
    def websub_getChannels(self, active=False):
        channels = []
        for channel in self.channels.find({'websub': True}, {'id': 1}):
            channels.append(channel['id'])
        return tuple(channels)
        
    def websub_getCallback(self, callbackId):
        return self.websub_callbacks.find_one({'id': callbackId})
        
    def websub_getCallbacks(self, channelId=None):
        callbacks = []
        
        if channelId:
            filter = {'channel': channelId}
        else:
            filter = {}
        
        for callback in self.websub_callbacks.find(filter, {'id': 1}):
            callbacks.append(callback['id'])
            
        return tuple(callbacks)
    
    def websub_savePost(self, callbackId, data):
        return self.websub_data.insert_one({'callback_id': callbackId, 'state': 'unprocessed', 'received_time': current_time(object=True), 'raw_data': data}).inserted_id
        
    def websub_getFirstPostData(self):
        data = self.websub_data.find_one({'state': 'unprocessed'}, {'_id': 1, 'raw_data': 1}, sort=[('received_time', 1)])
        
        if not data:
            return None
        
        self.websub_data.update_one({'_id': data['_id']}, {'$set': {'state': 'processing'}})
        
        return (data.get('_id'), data.get('raw_data'))
        
    def websub_deletePostProcessing(self, _id):
        self.websub_data.delete_one({'_id': _id})
        
    def websub_cleanRetired(self, days=1):
        days = self.datetime.utcnow() - self.timedelta(days=days)

        self.websub_callbacks.delete_many({'status': 'retired', 'retired_time': {'$lt': days}})
        self.websub_callbacks.delete_many({'status': 'requesting', 'requesting_time': {'$lt': days}})
        
        return True
        
    def websub_statistics(self):
        stats = {}
        
        stats['unprocessed_data'] = self.websub_data.count_documents({'state': 'unprocessed'})
        stats['active_callbacks'] = self.websub_callbacks.count_documents({'status': 'active'})
        
        return stats

    ##########################################
    #            QUEUE FUNCTIONS             #
    ##########################################
    
    def queue_newEndpoint(self, endpointId, description=''):
        self.queue_endpoints.insert_one({'id': endpointId, 'description': description, 'status': 'active', 'created_time': current_time(object=True)})
        return endpointId
    
    def queue_retireEndpoint(self, endpointId):
        return self.queue_endpoints.update_one({'id': endpointId}, {'$set': {'status': 'retired', 'retired_time': current_time(object=True)}})
    
    def queue_isActive(self, endpointId):
        status = self.queue_endpoints.find_one({'id': endpointId}, {'status': 1})

        if not status:
            return False
        elif status.get('status') == 'active':
            return True
        else:
            return False
            
    def queue_getEndpoints(self):
        return self.queue_endpoints.find({})
        
    def queue_cleanRetired(self, days=3):
        days = self.datetime.utcnow() - self.timedelta(days=days)

        self.queue_endpoints.delete_many({'status': 'retired', 'retired_time': {'$lt': days}})
        
        return True
        
    ##########################################
        
    def queue_insertQueue(self, videoId, endpointId=None):
        # if no document exists
        if not self.download_queue.count_documents({'id': videoId}) >= 1:
            self.download_queue.insert_one({'id': videoId, 'endpoint': endpointId, 'created_time': current_time(object=True), 'status': 'queued'}).inserted_id
            return True
            
        # key already in queue
        return False
        
    def queue_deleteQueue(self, videoId):
        if self.download_queue.delete_one({'id': videoId}):
            return True
        return False

    def queue_getQueue(self):
        return self.download_queue.find({})
        
    def queue_emptyQueue(self):
        return self.download_queue.delete_many({})

##########################################
#           HELPER FUNCTIONS             #
##########################################

def clean_info_json(originalInfo, format='dict'):

    unneeded_keys = ['automatic_captions', 'formats', 'thumbnails', 'thumbnail', 'subtitles', 'http_headers', 'url', '_filename', 'thumbnail']

    if type(originalInfo) != dict:
        try:
            originalInfo = json.loads(originalInfo)
        except json.decoder.JSONDecodeError as e:
            print('The given variable was not a valid dictionary. We tried to parse it anyway as JSON but that failed also!')
    
    for key in unneeded_keys:
        data.pop(key, None)
    
    if format == 'dict':
        return originalInfo
    elif format == 'str':
        return json.dumps(originalInfo)
    else:
        print('The requested output format is not supported!')
    
   
if __name__ == '__main__':
    mango = Mango('mongodb://root:example@192.168.66.140:27017')