class OIDC():
    """
    This function class is nothing more than a nonce and state store for security in the authentication mechanism.
    Additionally this class provides the function to generate redirect url's and check bearer tokens on their validity as well as caching jwt signing keys.
    Fairly barebones and should be 100% secure. (famous last words)
    This is made for form posted JWT's. While not the most secure it is the most easy way to implement. Moving on to a code based solution might be preferred in the future.
    """
    def __init__(self, app=None):
        self.states = {}
        self.nonces = {}
    
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        import requests
        import jwt

        config = app.config.copy()

        self.client_id = config['OIDC_ID']
        self.provider = config['OIDC_PROVIDER']
        self.domain = config['DOMAIN']
        self.window = 120  # the time window to allow states and nonces in seconds
        
        if self.provider[:8] != 'https://' or self.provider[-1] == '/':
            print('Incorrect OIDC provider URI', flush=True)
            exit()
            
        configuration = requests.get(f'{self.provider}/.well-known/openid-configuration').json()
        
        jwks_uri = configuration.get('jwks_uri')
        self.authorize_uri = configuration.get('authorization_endpoint')
        
        self.jwks_manager = jwt.PyJWKClient(jwks_uri)

    #######################################################

    def state_maintenance(self):
        from datetime import datetime
        
        pivot = datetime.now().timestamp() - self.window
        
        expired_states = [state for state, timestamp in self.states.items() if timestamp <= pivot]
        
        for state in expired_states:
            del self.states[state]

    def state_gen(self):
        import secrets
        from datetime import datetime
        
        self.state_maintenance()
        
        state = secrets.token_urlsafe(8)
        timestamp = datetime.now().timestamp()

        self.states[state] = timestamp

        return state
        
    def state_check(self, state):
        self.state_maintenance()

        if state in self.states:
            del self.states[state]
            return True
            
        return False
        
    #######################################################
    # Same code as above but a different store for nonces #
    #######################################################
    
    def nonce_maintenance(self):
        from datetime import datetime
        
        pivot = datetime.now().timestamp() - self.window
        
        expired_nonces = [nonce for nonce, timestamp in self.nonces.items() if timestamp <= pivot]
        
        for nonce in expired_nonces:
            del self.nonces[nonce]

    def nonce_gen(self):
        import secrets
        from datetime import datetime
        
        self.nonce_maintenance()
        
        nonce = secrets.token_urlsafe(8)
        timestamp = datetime.now().timestamp()

        self.nonces[nonce] = timestamp

        return nonce
        
    def nonce_check(self, nonce):
        self.nonce_maintenance()

        if nonce in self.nonces:
            del self.nonces[nonce]
            return True
            
        return False
    
    #######################################################
    
    def generate_redirect(self):
        return str(f'{self.authorize_uri}'
                   '?response_mode=form_post&response_type=id_token&scope=openid'
                   f'&redirect_uri={self.domain}/auth/callback'
                   f'&client_id={self.client_id}'
                   f'&nonce={self.nonce_gen()}'
                   f'&state={self.state_gen()}')
        
    def check_bearer(self, token):
        import jwt

        try:
            signing_key = self.jwks_manager.get_signing_key_from_jwt(token).key
            decoded = jwt.decode(token, signing_key,
                                 algorithms=jwt.algorithms.get_default_algorithms(),
                                 issuer=self.provider,
                                 require=['aud', 'client_id', 'exp', 'iat', 'iss', 'rat', 'sub'],
                                 audience=self.client_id,
                                 leeway=5)
        except Exception as e:
            print(e, flush=True)
            return False

        # double check if given token is really requested by us
        if not self.nonce_check(decoded.get('nonce', None)):
            return False

        return decoded.get('sub', False)
