from celery import shared_task
from flask import current_app

##########################################
#              CELERY TASKS              #
##########################################

@shared_task()
def websub_subscribe_callback(channelId):
    import requests
    from .nosql import get_nosql
    
    # check if a callback already exists for channel
    answer = get_nosql().websub_existsCallback(channelId, channel=True)
    
    if not answer:
        callbackId = get_nosql().websub_newCallback(channelId)
    else:
        callbackId = answer
    
    url = 'https://pubsubhubbub.appspot.com/subscribe'
    data = {
        'hub.callback': f'{current_app.config["DOMAIN"]}/api/websub/{callbackId}',
        'hub.topic': f'https://www.youtube.com/xml/feeds/videos.xml?channel_id={channelId}',
        'hub.verify': 'async',
        'hub.mode': 'subscribe',
        'hub.verify_token': '',
        'hub.secret': '',
        'hub.lease_numbers': '432000',
    }
    
    get_nosql().websub_requestingCallback(callbackId)
    response = requests.post(url, data=data)
    if response.status_code == 202:
        return True
    
    # maybe handle errors?
    
    return False
    
@shared_task()
def websub_unsubscribe_callback(callbackId):
    import requests
    from .nosql import get_nosql
    
    answer = get_nosql().websub_existsCallback(callbackId)
    
    if not answer:
        return False
        
    channelId = get_nosql().websub_getCallback(callbackId).get('channel')
    
    url = 'https://pubsubhubbub.appspot.com/subscribe'
    data = {'hub.callback': f'{current_app.config["DOMAIN"]}/api/websub/{callbackId}',
            'hub.topic': f'https://www.youtube.com/xml/feeds/videos.xml?channel_id={channelId}',
            'hub.verify': 'async',
            'hub.mode': 'unsubscribe'
           }
    
    get_nosql().websub_retiringCallback(callbackId)
    response = requests.post(url, data=data)
    
    if response.status_code == 202:
        return True
        
    # maybe handle errors?
        
    return False
    
@shared_task()
def websub_process_data():
    from .nosql import get_nosql
    
    while True:
        blob = get_nosql().websub_getFirstPostData()
        if not blob:
            break
        
        _id, data = blob
        
        parsed = do_parse_data(data)
        if parsed:
            state, channelId, videoId = parsed
        
            if state == 'added':
                if not get_nosql().check_exists(videoId):  # if video not exists
                   get_nosql().queue_insertQueue(videoId, 'WebSub')

            elif state == 'removed':
                pass
            
        get_nosql().websub_deletePostProcessing(_id)

@shared_task()
def websub_renew_expiring(hours=6):
    from .nosql import get_nosql
    from datetime import datetime, timedelta
    
    count = 0

    for callbackId in get_nosql().websub_getCallbacks():
        data = get_nosql().websub_getCallback(callbackId)
        
        if data.get('status') not in ['active']:  # callback not active
            continue
        
        pivot = datetime.utcnow() + timedelta(hours=hours)  # hours past now
        expires = data.get('activation_time') + timedelta(seconds=data.get('lease'))  # callback expires at
        
        if pivot <= expires:  # expiration happens after n hours fron now
            continue  # skip callback
        
        # expiration happens within n hours
        websub_subscribe_callback.delay(data.get('channel'))
        
        # limit amount of subscribe requests to spread out the requests over time
        count = count + 1
        if count >= 100:
            break

##########################################
#              TASK MODULES              #
##########################################

def do_parse_data(data):
    import xml.etree.ElementTree as ET
    
    data = data.decode('utf-8')

    try:
        root = ET.fromstring(data)
    except ET.ParseError:
        print('Not XML')
        return False

    yt = any(child.tag.startswith('{http://www.youtube.com/xml/schemas/2015}') for child in root.iter())
    at = any(child.tag.startswith('{http://purl.org/atompub/tombstones/1.0}') for child in root.iter())

    if yt and not at:
        # Video published
        state = 'added'
        ns = {'yt': 'http://www.youtube.com/xml/schemas/2015', '': 'http://www.w3.org/2005/Atom'}
        entry = root.find('.//{http://www.w3.org/2005/Atom}entry')
        videoId = entry.find('./yt:videoId', ns).text
        channelId = entry.find('./yt:channelId', ns).text
    elif not yt and at:
        # Video hidden
        state = 'removed'
        ns = {'at': 'http://purl.org/atompub/tombstones/1.0', '': 'http://www.w3.org/2005/Atom'}
        deleted_entry = root.find('.//{http://purl.org/atompub/tombstones/1.0}deleted-entry')
        videoId = deleted_entry.attrib['ref'].split(':')[-1]
        channelId = deleted_entry.find('./at:by/uri', ns).text.split('/')[-1]
    else:
        print('Unknown xml')
        return False
        
    return (state, channelId, videoId)